
#using models found in Statistical Analysis of Network Data with R (Kolacyzk & Csardi) https://kateto.net/netscix2016.html
library(MASS)
library(mixtools)
library(igraph)
library(iGraphMatch)
library(class)
library(factoextra)
library(tree)
library(randomForest)
library(caret)
library(e1071)
library(mclust)
library(irlba)
library(ggplot2)
library(tidyverse)
library(socviz)
library(Matrix)
library(flexclust)
library(igraph)
library(iGraphMatch)
library(MESS)
library(remotes)
library(ggplot2)
library(reshape2)
library(viridis)
library(colorspace)
library(extrafont)

TS2 <- function(A,B){
  return(norm(A-B, "F")^2)
}

shuff_ord <- function(x,vec1, vec2) 
  #function that swaps * number of vertices from block 1 and 2
{
  temp <- vec2
  x[vec2] <- vec1
  x[vec1] <- temp
  return(x)
} 

ase <- function(A, dim) {
  require(irlba)
  
  #diag(A) <- rowSums(A) / (nrow(A)-1) # diagaug line
  
  if(nrow(A) >= 1000){
    A.svd <- irlba(A, nu = dim, nv = dim)
    A.svd.values <- A.svd$d[1:dim]
    A.svd.vectors <- A.svd$v[,1:dim,drop=F]
    if(dim == 1)
      A.coords <- sqrt(A.svd.values) * A.svd.vectors
    else
      A.coords <- A.svd.vectors %*% diag(sqrt(A.svd.values))
  } else{
    A.svd <- svd(A)
    A.svd.values <- A.svd$d[1:dim]
    if(dim == 1)
      A.coords <- matrix(A.svd$v[,1,drop=F] * sqrt(A.svd$d[1]),ncol=dim)
    else
      A.coords <- A.svd$v[,1:dim] %*% diag(sqrt(A.svd$d[1:dim]))
  }
  
  return(list(X=A.coords,D=A.svd.values))
}

shuff_perm <- function( n, k ){
  #  generate a random permutation of {1,2,3,...,n} that 
  #  shuffles exactly k labels
  #  
  #  n is assumed an integer > 0, with k in {1,2,3,...,n}
  #  ch is a variable we introduce as a test check
  
  ch <- 0  
  while( ch == 0 ){
    # generate a random candidate permutation
    x <- sort(sample(n,k))
    y <- sample(x)
    cand <- c(1:n)
    cand[x] <- y
    # compute number of fixed points in the shuffled verts
    if( sum( x == y) == 0){
      ch <- 1 }
  }
  # output permutation sigma in matrix form where row 1 is 1:n and
  # row 2 entry i is the value sigma(i)
  return(matrix(c(1:n,cand),nrow=2,byrow=T))
}


make_Omni<- function(A,w)
{
  n<-dim(A[[1]])[1]
  m<-length(A)
  ind<-list()
  for(i in 1:m){
    ind[[i]]<-c( (n*(i-1)+1): (n*i))
  }
  O<-matrix(0,n*m,n*m)
  for(i in 1:m){ 
    for(j in 1:m){
      O[ ind[[i]] , ind[[j]] ]<-as.matrix((w[i]*A[[i]]+w[j]*A[[j]])/(w[i]+w[j]))
    }}
  for(i in 1:m){O[ ind[[i]] , ind[[i]] ]<-as.matrix(A[[i]])}
  return(O)
}

procrustes<-function(X,Y, type = "I"){
  if(type == "C"){
    X <- X/norm(X, type = "F")*sqrt(nrow(X))
    Y <- Y/norm(Y, type = "F")*sqrt(nrow(Y))
  }
  if(type == "D"){
    tX <- rowSums(X^2)
    tX[tX <= 1e-15] <- 1
    tY <- rowSums(Y^2)
    tY[tY <= 1e-15] <- 1
    X <- X/sqrt(tX)
    Y <- Y/sqrt(tY)
  }
  
  tmp <- t(X) %*% Y
  tmp.svd <- svd(tmp)
  W <- tmp.svd$u %*% t(tmp.svd$v)
  return(list(diff = norm(X%*%W - Y, type = "F"), W = W))
}

buildOmni <- function(Alist)
{
  # elements of A are assumed symmetric!
  require(Matrix)
  
  m <- length(Alist)
  nvec <- sapply(Alist, nrow)
  nsum <- c(0, cumsum(nvec))
  
  omni <- as.matrix(bdiag(Alist))
  for (i in 1:(m-1)) {
    irng <- (nsum[i]+1):nsum[i+1]
    for (j in (i+1):m) {
      jrng <- (nsum[j]+1):nsum[j+1]
      omni[irng,jrng] <- (as.matrix(Alist[[i]]) + as.matrix(Alist[[j]]))
    }
  }
  omni <- (omni + t(omni)) / 2
  return(omni)
}

getElbows <- function(dat, n = 3, threshold = FALSE, plot = F) {
  if (is.matrix(dat))
    d <- sort(apply(dat,2,sd), decreasing=T)
  else
    d <- sort(dat,decreasing=TRUE)
  
  if (!is.logical(threshold))
    d <- d[d > threshold]
  
  p <- length(d)
  if (p == 0)
    stop(paste("d must have elements that are larger than the threshold ",
               threshold), "!", sep="")
  
  lq <- rep(0.0, p)                     # log likelihood, function of q
  for (q in 1:p) {
    mu1 <- mean(d[1:q])
    mu2 <- mean(d[-(1:q)])              # = NaN when q = p
    sigma2 <- (sum((d[1:q] - mu1)^2) + sum((d[-(1:q)] - mu2)^2)) / (p - 1 - (q < p))
    lq[q] <- sum( dnorm(  d[1:q ], mu1, sqrt(sigma2), log=TRUE) ) +
      sum( dnorm(d[-(1:q)], mu2, sqrt(sigma2), log=TRUE) )
  }
  
  q <- which.max(lq)
  if (n > 1 && q < p) {
    q <- c(q, q + getElbows(d[(q+1):p], n=n-1, plot=FALSE))
  }
  
  if (plot==TRUE) {
    if (is.matrix(dat)) {
      sdv <- d # apply(dat,2,sd)
      plot(sdv,type="b",xlab="dim",ylab="stdev")
      points(q,sdv[q],col=2,pch=19)
    } else {
      plot(dat, type="b")
      points(q,dat[q],col=2,pch=19)
    }
  }
  
  return(q)
}

scramble <- function(x, k) 
{
  if(k == 0)
  {
    return(x)
  }
  else
  {
    sum <- 0
    while(sum != k)
    {
      y <- x
      new_ord <- sample(x,k)
      y[sample(new_ord)] <- x[new_ord]
      sum <- sum(x!=y)
    }
    return(y)
  } 
}

sgm<-function(A,B,m,start,iteration){
  #seeds are assumed to be vertices 1:m in both graphs
  require('clue')
  totv1<-ncol(A)
  totv2<-ncol(B)
  
  #If A is bigger than B, fill B with 0's until size of A
  if(totv1>totv2)
  {
    A[A==0]<- -1
    B[B==0]<- -1
    diff<-totv1-totv2
    
    for (j in 1:diff)
    {
      B<-cbind(rbind(B,0),0)
    }
  }
  #If B is bigger than A, fill A with 0's until size of B
  else if(totv1<totv2)
  {
    A[A==0]<- -1
    B[B==0]<- -1
    diff<-totv2-totv1
    
    for (j in 1:diff)
    {
      A<-cbind(rbind(A,0),0)
    }
  } 
  
  totv<-max(totv1,totv2)
  
  n<-totv-m
  
  perm<-sample(n)
  
  ####BREAKS RIGHT HERE????#######
  Z<-diag(n)[perm,]
  
  if (m != 0)
  {
    A12<-rbind(A[1:m,(m+1):(m+n)])
    A21<-cbind(A[(m+1):(m+n),1:m])
    B12<-rbind(B[1:m,(m+1):(m+n)])
    B12<-B12[,perm]
    B21<-cbind(B[(m+1):(m+n),1:m])
    B21<-B21[perm,]
  }
  
  if (m==0)
  {
    A12<-matrix(0,n,n)
    A21<-matrix(0,n,n)
    B12<-matrix(0,n,n)
    B21<-matrix(0,n,n)
  }
  
  if (n==1)
  { 
    A12=t(A12)
    A21=t(A21)
    B12=t(B12)
    B21=t(B21)
  }
  A22<-A[(m+1):(m+n),(m+1):(m+n)]
  B22<-B[(m+1):(m+n),(m+1):(m+n)]
  patience<-iteration
  tol<-1
  B22<-B22[perm,perm]
  P<-start[,perm]
  toggle<-1
  iter<-0
  x<- A21%*%t(B21)
  y<- t(A12)%*%B12
  
  while (toggle==1 & iter<patience)
  {
    iter<-iter+1
    z<- A22%*%P%*%t(B22)
    w<- t(A22)%*%P%*%B22
    Grad<-x+y+z+w;
    mm=max(abs(Grad))
    ind<-matrix(solve_LSAP(Grad+matrix(mm,totv-m,totv-m), maximum =TRUE))
    T<-diag(n)
    T<-T[ind,]
    wt<-t(A22)%*%T%*%B22
    c<-sum(diag(w%*%t(P)))
    d<-sum(diag(wt%*%t(P)))+sum(diag(w%*%t(T)))
    e<-sum(diag(wt%*%t(T)))
    u<-sum(diag(t(P)%*%x+t(P)%*%y))
    v<-sum(diag(t(T)%*%x+t(T)%*%y))
    
    if( c-d+e==0 && d-2*e+u-v==0)
    {
      alpha<-0
    }
    else
    {
      alpha<- -(d-2*e+u-v)/(2*(c-d+e))
    }
    
    f0<-0
    
    f1<- c-e+u-v
    falpha<-(c-d+e)*alpha^2+(d-2*e+u-v)*alpha
    
    if(alpha < tol && alpha > 0 && falpha > f0 && falpha > f1)
    {
      P<- alpha*P+(1-alpha)*T
    }
    else if(f0 > f1)
    {
      P<-T
    }
    else
    {
      toggle<-0}
  }
  
  D<-P
  corr<-matrix(solve_LSAP(P, maximum = TRUE))
  
  P=diag(n)
  P=rbind(cbind(diag(m),matrix(0,m,n)),cbind(matrix(0,n,m),P[corr,]))
  Z=rbind(cbind(diag(m),matrix(0,m,n)),cbind(matrix(0,n,m),Z))
  
  corr<-cbind(matrix((m+1):totv, n),matrix(m+corr,n))
  
  corr<-matrix(solve_LSAP(P%*%Z, maximum = TRUE))
  
  if(m==0)
  {
    return(list(corr=corr, P=P%*%Z, D=D%*%Z))
  }
  else
  {
    return(list(corr=corr, P=P[c(1:m,perm+m),], D=D[perm,]))
  }
  
}

derange<-function(n){
  # returns a derangement of 1:n
  c=0
  while( c==0 ){d<-sample(n);
  if(sum(d==1:n)==0){c=1}}
  return(d)
}


set.seed(12)
#######Brain Code:

g1 = read_graph(url("http://www.cis.jhu.edu/~parky/Microsoft/JHU-MSR/ZMx2/BNU1/DS01216/subj1-scan1.graphml"),format="graphml")
A1 <- (g1[] +t(g1[]))/2
A1[A1>0] <- 1

g2 = read_graph(url("http://www.cis.jhu.edu/~parky/Microsoft/JHU-MSR/ZMx2/BNU1/DS01216/subj1-scan2.graphml"),format="graphml")
A2 <- (g2[] +t(g2[]))/2
A2[A2>0] <- 1

g3 = read_graph(url("http://www.cis.jhu.edu/~parky/Microsoft/JHU-MSR/ZMx2/BNU1/DS01216/subj2-scan1.graphml"),format="graphml")
B1 <- (g3[] +t(g3[]))/2
B1[B1>0] <- 1

g4 = read_graph(url("http://www.cis.jhu.edu/~parky/Microsoft/JHU-MSR/ZMx2/BNU1/DS01216/subj2-scan2.graphml"),format="graphml")
B2 <- (g4[] +t(g4[]))/2
B2[B2>0] <- 1

cv <- intersect( intersect(V(g1)$name,V(g2)$name ), V(g3)$name)



A1<-A1[cv,cv]
ee<-irlba(A1,30)
da1<-getElbows(ee$d)[2]

A2<-A2[cv,cv]
ee<-irlba(A2,30)
da2<-getElbows(ee$d)[2]

B1<-B1[cv,cv]
ee<-irlba(B1,30)
db1<-getElbows(ee$d)[2]

#B2<-B2[cv,cv]
#ee<-irlba(B2,30)
#db2<-getElbows(ee$d)[2]

d <- max(da1, da2, db1)
X1 <- ase(A1,d)$X
X2 <- ase(A2,d)$X
Y1 <- ase(B1,d)$X
#Y2 <- ase(B2,d)$X

nn<-dim(A1)[1]
ord <- 1:nn

PPP2a_Sep1 <- matrix(0,6,6)
PPP2n_Sep1<- matrix(0,6,6)
PPP2a_unshuff_Sep1 <- matrix(0,6,6)
PPP2n_unshuff_Sep1<- matrix(0,6,6)


nMC = 5 #number of repetitions MC
times <- 200 #number of bootstraps 
for(eee in 1:nMC)
{
  perm<-sample(nn)
  X1<-X1[perm,]
  X2<-X2[perm,]
  Y1<-Y1[perm,]
  
  k <- c(0, 50, 150, 200, 250, 350)
  PP <- list()
  for(ii in 1:6){
    if(k[ii] != 0){
      pp <- derange(k[ii])
      pp <- c(pp,(k[ii]+1):nn)
      PP[[ii]] <- pp}else{PP[[ii]]<-c(1:nn)}
  }
  crit_ts2 <- rep(0,6)
  crit_ts2_unshuff <- rep(0,6)
  for(iii in 1:6)
  {
    #bootstrap null critical value
    t2 <- rep(0,times)
    t2unshuff <- rep(0,times)
    for(j in 1:times)
    {
      AA1 <- sample_dot_product(t(X1),directed=FALSE)[]
      XX1<-ase(AA1,d)$X
      Phat1<-XX1%*%t(XX1)
      
      AA2 <- sample_dot_product(t(X2),directed=FALSE)[]
      AA2shuff<-AA2[PP[[iii]], PP[[iii]]]
      XX2<-ase(AA2shuff[],d)$X
      Phat2shuff<-XX2%*%t(XX2)
      
      t2[j]<-TS2(Phat1,Phat2shuff)
      Mk<-c((k[iii]+1):nn)
      
      if(k[iii]!=0)
      {
        unshuffperm <- gm(Phat1, Phat2shuff, seeds = Mk)
        t2unshuff[j]<-norm(as.matrix(Phat1 - Phat2shuff[unshuffperm$corr_B,unshuffperm$corr_B]), "F")^2
      }
      else
      {
        t2unshuff[j]<-t2[j]
      }
      
    }
    crit_ts2[iii]<-quantile(t2,0.95)
    crit_ts2_unshuff[iii]<-quantile(t2unshuff,0.95)
  }
  
  alt_ts2a <- matrix(0,6,times)
  alt_ts2n <- matrix(0,6,times)
  alt_ts2a_unshuff <- matrix(0,6,times)
  alt_ts2n_unshuff <- matrix(0,6,times)
  for(iii in 1:6)
  {
    #bootstrap for power
    for(j in 1:times)
    {
      AA1 <- sample_dot_product(t(X1),directed=FALSE)[]
      XX1<-ase(AA1,d)$X
      Phat1<-XX1%*%t(XX1)
      
      AA2 <- sample_dot_product(t(X2),directed=FALSE)[]
      XX2<-ase(AA2[PP[[iii]], PP[[iii]]],d)$X
      Phat2shuff<-XX2%*%t(XX2)
      
      BB1 <- sample_dot_product(t(Y1),directed=FALSE)[]
      YY1shuff<-ase(BB1[PP[[iii]], PP[[iii]]],d)$X
      Phat3shuff<-YY1shuff%*%t(YY1shuff)
      
      alt_ts2a[iii,j]<-TS2(Phat1, Phat3shuff)
      alt_ts2n[iii,j]<-TS2(Phat1, Phat2shuff)
      
      if(k[iii]!=0)
      {
        Mk<-c((k[iii]+1):nn)
        unshuffperm <- gm(Phat1, Phat2shuff, seeds = Mk)
        alt_ts2n_unshuff[iii,j]<-TS2(Phat1,Phat2shuff[unshuffperm$corr_B,unshuffperm$corr_B])
        unshuffperm <- gm(Phat1, Phat3shuff, seeds = Mk)
        alt_ts2a_unshuff[iii,j]<-TS2(Phat1,Phat3shuff[unshuffperm$corr_B,unshuffperm$corr_B])
      }
      else
      {
        alt_ts2n_unshuff[iii,j]<-alt_ts2n[iii,j]
        alt_ts2a_unshuff[iii,j]<-alt_ts2a[iii,j]
      }
      
    }
  }
  
  Pow2a <- matrix(NA,6,6 )
  Pow2n <- matrix(NA,6,6 )
  Pow2a_unshuff <- matrix(NA,6,6 )
  Pow2n_unshuff <- matrix(NA,6,6 )
  for(i in 1:6)
  {
    for(j in 1:i)
    {
      Pow2a[i,j] <- sum( alt_ts2a[j,] > crit_ts2[i])/times 
      Pow2n[i,j] <- sum( alt_ts2n[j,] > crit_ts2[i])/times
      Pow2a_unshuff[i,j] <- sum( alt_ts2a_unshuff[j,] > crit_ts2_unshuff[i])/times 
      Pow2n_unshuff[i,j] <- sum( alt_ts2n_unshuff[j,] > crit_ts2_unshuff[i])/times
    }
  }
  PPP2a_Sep1 <-PPP2a_Sep1+Pow2a
  PPP2n_Sep1 <-PPP2n_Sep1+Pow2n
  PPP2a_unshuff_Sep1 <-PPP2a_unshuff_Sep1+Pow2a_unshuff
  PPP2n_unshuff_Sep1 <-PPP2n_unshuff_Sep1+Pow2n_unshuff
}

save(PPP2a_Sep1, file="PPP2aSeparate8.RData")
save(PPP2n_Sep1, file="PPP2nSeparate8.RData")
save(PPP2a_unshuff_Sep1, file="PPP2a_unshuffSeparate8.RData")
save(PPP2n_unshuff_Sep1, file="PPP2n_unshuffSeparate8.RData")